import logging
import sys

import loguru

logger = loguru.logger


class LoguruHandler(logging.Handler):  # pragma: no cover
	"""将 logging 的日志转发到 loguru。"""

	def emit(self, record: logging.LogRecord):
		try:
			level = logger.level(record.levelname).name
		except ValueError:
			level = record.levelno

		frame, depth = sys._getframe(6), 6
		while frame and frame.f_code.co_filename == logging.__file__:
			frame = frame.f_back
			depth += 1

		logger.opt(depth=depth, exception=record.exc_info).log(
			level, record.getMessage()
		)



from httpx import AsyncClient

MAX_CONCURRENT = 10

header = {
	'user-agent': r'Mozilla/5.0 (Linux; Android 6.0.1; RIDGE 4G Build/LRX22G) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2887.55 Mobile Safari/537.36',
	'referer': r'https://sp.61.com',
}

client = AsyncClient(headers=header)



from abc import ABC, abstractmethod
from collections.abc import Callable, Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar

from UnityPy.enums import ClassIDType
from UnityPy.files.ObjectReader import ObjectReader
from UnityPy.tools.extractor import EXPORT_TYPES

from albi0.typing import (
	ExportHandlerResult,
	ObjectPath,
	PathTypes,
	T_AssetPostHandler,
	T_ExportHandler,
	T_ObjPreHandler,
)

if TYPE_CHECKING:
	from UnityPy.classes import Object
	from UnityPy.files import ObjectReader

KT = TypeVar('KT')
T_Callable = TypeVar('T_Callable', bound=Callable)


class SkipCurrentHandlerGroup(Exception): ...


class StopExtractThisObject(Exception): ...


class BaseHandlerGroup(ABC):
	@abstractmethod
	def register(self, *args, **kwargs):
		raise NotImplementedError


class ListHandlerGroup(list[T_Callable], BaseHandlerGroup):
	def register(self):
		def _decorator(func: T_Callable) -> T_Callable:
			self.append(func)
			return func

		return _decorator


class DictHandlerGroup(dict[KT, T_Callable], BaseHandlerGroup):
	def register(self, key: KT):
		def _decorator(func: T_Callable) -> T_Callable:
			self[key] = func
			return func

		return _decorator


class MultiValuesDictHandlerGroup(dict[KT, list[T_Callable]], BaseHandlerGroup):
	def register(self, key: KT):
		def _decorator(func: T_Callable) -> T_Callable:
			try:
				self[key].append(func)
			except KeyError:
				self[key] = [func]
			return func

		return _decorator


class AssetPostHandlerGroup(ListHandlerGroup[T_AssetPostHandler]):
	"""Asset后处理程序组"""

	def handle(
		self, obj: 'ObjectReader', export_dir: Path
	) -> tuple[ObjectReader[Any], Path]:
		for handler in self:
			try:
				obj, export_dir = handler(obj, export_dir)
			except SkipCurrentHandlerGroup:
				break

		return obj, export_dir


class ObjPreHandlerGroup(MultiValuesDictHandlerGroup[ClassIDType, T_ObjPreHandler]):
	"""
	Unity对象预处理程序组，key为Unity对象ClassID，value为handler列表。
	处理方式为按ClassID链式处理。
	"""

	def handle(
		self, obj: 'Object', obj_path: ObjectPath
	) -> tuple['Object', ObjectPath]:
		if (obj_reader := obj.object_reader) is None:
			raise ValueError(f'{obj_path} has no object reader')
		handlers = self.get(obj_reader.type, [])
		for handler in handlers:
			try:
				obj, obj_path = handler(obj, obj_path)
			except SkipCurrentHandlerGroup:
				break

		return obj, obj_path


class ExportHandlerGroup(DictHandlerGroup[ClassIDType, T_ExportHandler]):
	"""
	导出处理器组，用于自定义Unity对象的导出逻辑，
	key为Unity对象ClassID，value为导出函数。

	当没有对应的handler时，会使用UnityPy的默认导出逻辑。
	"""

	def __init__(self, mapping: Mapping[ClassIDType, T_ExportHandler] | None = None):
		super().__init__({**EXPORT_TYPES, **(mapping or {})})

	def handle(
		self,
		obj: 'Object',
		obj_type: ClassIDType,
		export_filename: PathTypes,
		*,
		suffix: str,
		export_unknown_as_typetree: bool = False,
	) -> ExportHandlerResult | None:
		export_filename = Path(export_filename)
		try:
			handler = self[obj_type]
		except KeyError:
			if export_unknown_as_typetree:
				handler = self[ClassIDType.MonoBehaviour]
			else:
				return None

		return handler(obj, export_filename, suffix)



from collections.abc import Callable
import os
from pathlib import Path, PurePath
from typing import NewType, Optional, TypeAlias, TypeVar

from UnityPy.classes import NamedObject
from UnityPy.files import ObjectReader, SerializedFile

PathTypes: TypeAlias = Path | str

DecryptionMethod: TypeAlias = Callable[[memoryview], memoryview]

DownloadPostProcessMethod = Callable[[bytes], bytes]

ObjectPath = NewType('ObjectPath', PurePath)

T_PathLike = TypeVar('T_PathLike', str, ObjectPath, os.PathLike)

T_NamedObject = TypeVar('T_NamedObject', bound='NamedObject')

T_AssetPostHandler: TypeAlias = Callable[
	['ObjectReader', Path], tuple['ObjectReader', Path]
]
T_ObjPreHandler: TypeAlias = Callable[
	[T_NamedObject, ObjectPath], tuple[T_NamedObject, ObjectPath]
]
ExportHandlerResult: TypeAlias = list[tuple[Optional['SerializedFile'], int]]
T_ExportHandler: TypeAlias = Callable[
	[T_NamedObject, PathTypes, str], ExportHandlerResult
]



import importlib
from pathlib import Path

from asyncer import syncify
import click

from albi0.request import client as httpx_client


@click.group(invoke_without_command=True)
@click.version_option(
	None,
	'-V',
	'--version',
	prog_name='albi0',
)
@click.option(
	'-d',
	'--cwd',
	default=None,
	help='The working directory.',
	type=Path,
	is_eager=True,
	expose_value=False,
)
@click.pass_context
@syncify
async def cli(ctx: click.Context):
	importlib.import_module('albi0.plugins')

	ctx.call_on_close(on_close)
	if ctx.invoked_subcommand is not None:
		return


@syncify
async def on_close():
	await httpx_client.aclose()


from .commands import extract, list_, update

cli.add_command(update)
cli.add_command(extract)
cli.add_command(list_)



import glob
import itertools
from pathlib import Path

from asyncer import syncify
import click

from albi0.extract.extractor import extractors
from albi0.log import logger
from albi0.utils import join_path, timer


@click.command(context_settings={'ignore_unknown_options': True})
@click.option(
	'-o',
	'--output-dir',
	default=None,
	type=click.Path(file_okay=False, writable=True),
)
@click.option(
	'-n',
	'--extractor-name',
	default='default',
	type=str,
)
@click.option(
	'-e',
	'--export-as-is',
	default=False,
	is_flag=True,
	show_default=True,
)
@click.option(
	'-m',
	'--merge-extract',
	default=False,
	is_flag=True,
	show_default=True,
)
@click.option(
	'-t',
	'--parallel-threads',
	'--max-workers',
	default=4,
	type=int,
	show_default=True,
)
@click.argument('patterns', nargs=-1, default=None)
@click.pass_context
@syncify
async def extract(
	ctx: click.Context,
	patterns: list[str] | None,
	output_dir: str | None,
	extractor_name: str,
	export_as_is: bool,
	merge_extract: bool,
	parallel_threads: int,
):
	output_dir = output_dir or '.'
	patterns = patterns or []
	ab_paths: set[Path] = {
		Path(path).resolve()
		for path in itertools.chain.from_iterable(map(glob.iglob, patterns))
	}

	if export_as_is:
		extractor_name = 'default'

	extractor_set = extractors.get_processors(extractor_name)

	if not extractor_set:
		click.echo(f'找不到输入的提取器/组{extractor_name}')
		return

	with timer('✅ 提取完成~ 总耗时: {duration:.2f}s'):
		for extractor in extractor_set:
			click.echo(f'运行提取器：{extractor.name}')
			try:
				extractor.extract_asset(
					*ab_paths,
					export_dir=join_path(extractor_name, output_dir),
					merge_extract=merge_extract,
					max_workers=parallel_threads,
				)
				click.echo(f'✅ {extractor.name}提取完成~')
			except Exception as e:
				logger.opt(exception=e).error(e)
				click.echo(f'❌ {extractor.name}提取失败.')
				raise



from typing import Protocol, TypeVar


class ProcessorProtocol(Protocol):
	name: str
	desc: str


T_Processor = TypeVar('T_Processor', bound=ProcessorProtocol)


class ProcessorContainer(dict[str, T_Processor]):
	"""Processor容器，用于存储Updater、Extractor等符合ProcessorProtocol协议的类

	提供了使用'.'简易分组功能，
	例如::

		from albi0.container import ProcessorContainer

		container = ProcessorContainer()
		foo = FooUpdater()
		bar = BarUpdater()
		container['newseer.Foo'] = foo
		container['newseer.Bar'] = bar
		assert container.get_by_group('newseer') == {foo, bar}
	"""

	def get_by_group(self, group_name: str) -> set[T_Processor]:
		"""根据组名获取整组Processor"""
		processors = {v for k, v in self.items() if k.split('.')[0] == group_name}
		return processors

	def get_processors(self, name: str) -> set[T_Processor]:
		"""如果传入的参数为组名则获取整组Processor，否则获取单个"""
		return {processor} if (processor := self.get(name)) else self.get_by_group(name)



from contextlib import suppress
import gzip
from pathlib import Path
from typing import TYPE_CHECKING

import orjson
from slpp import slpp
from UnityPy.classes import PPtr
from UnityPy.enums.ClassIDType import ClassIDType

from albi0.extract import Extractor
from albi0.extract.registry import (
	AssetPostHandlerGroup,
	ObjPreHandlerGroup,
	StopExtractThisObject,
)
from albi0.typing import ObjectPath
from albi0.update import Downloader, Updater
from albi0.update.version import LocalFileName, Manifest, ManifestItem
from albi0.updaters import YooVersionManager
from albi0.updaters.yoo_version_manager import PackageManifest
from albi0.utils import join_path, join_url, remove_all_suffixes

if TYPE_CHECKING:
	from UnityPy.classes import TextAsset, Texture2D
	from UnityPy.files.ObjectReader import ObjectReader


def load_lua_table(data: str):
	"""加载lua数据为Python object"""
	start = data.find('{')
	return slpp.decode(data[start:] if start != -1 else data)


def default_decryption_method(data: memoryview):
	if data[:32] == b'\00' * 32:
		return data[32:]

	return data


class SeerProjectVersionManager(YooVersionManager):
	def _simplify_manifest(self, data: PackageManifest) -> Manifest:
		version = data['PackageVersion']
		items = {}
		for item in data['BundleList']:
			local_basename = item['BundleName']
			remote_filehash = item['FileHash']
			local_fn = LocalFileName(join_path(self.local_path, local_basename))
			items[local_fn] = ManifestItem(
				f'{join_url(self.remote_path, remote_filehash)}.bundle',
				f'{local_basename}.bundle',
				remote_filehash.encode(),
			)
		return Manifest(version=version, items=items)


obj_pre = ObjPreHandlerGroup()
asset_post = AssetPostHandlerGroup()
Extractor(
	'seerproject',
	'赛尔计划资源提取器',
	decryption_method=default_decryption_method,
	asset_posthandler_group=asset_post,
	obj_prehandler_group=obj_pre,
)


@asset_post.register()
def non_bundle_package_handler(obj: 'ObjectReader', export_dir: Path):
	assets_file = obj.assets_file
	container = assets_file.container
	if (
		len(values := container.values()) == 1
		and isinstance(values[0], PPtr)
		and Path(filename := container.keys()[0]).suffix in ('.unity', '.prefab')
	):
		export_filename: Path = Path(export_dir, filename)
		export_filename.parent.mkdir(parents=True, exist_ok=True)
		export_filename.write_bytes(assets_file.save(packer='original'))
		raise StopExtractThisObject

	return obj, export_dir


@obj_pre.register(ClassIDType.TextAsset)
def textasset_prehandler(
	obj: 'TextAsset', obj_path: ObjectPath
) -> tuple['TextAsset', ObjectPath]:
	suffix = '.txt'
	if len(suffixes := obj_path.suffixes):
		if suffixes[0] in ('.atlas', '.skel'):
			suffix = suffixes[0]

	script = obj.m_Script.encode('utf-8', 'surrogateescape')
	with suppress(gzip.BadGzipFile):
		script = gzip.decompress(script)

	if obj_path.is_relative_to('Assets/Game/Lua/'):
		suffix = '.lua'
		text = script.decode()

		if obj_path.parts[-2] == 'data':
			suffix = '.json'
			lua_table = ''.join(text.splitlines()[:-1])  # 删除最后一行
			script = orjson.dumps(
				load_lua_table(lua_table),
				option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS,
			)

	obj.m_Script = script.decode('utf-8', 'surrogateescape')

	obj_path = remove_all_suffixes(obj_path).with_suffix(suffix)
	return obj, obj_path


@obj_pre.register(ClassIDType.Sprite)
@obj_pre.register(ClassIDType.Texture2D)
def texture2d_prehandler(
	obj: 'Texture2D', obj_path: ObjectPath
) -> tuple['Texture2D', ObjectPath]:
	if obj.image.mode == 'RGBA' and obj_path.suffix != '.png':
		obj_path = obj_path.with_suffix('.png')
	return obj, obj_path


Updater(
	'seerproject.ab',
	'赛尔计划AB包下载器',
	version_manager=SeerProjectVersionManager(
		'SpPackage',
		remote_path='https://sp.61.com/source/taomee/Android/',
		local_path='./seerproject/assetbundles',
	),
	downloader=Downloader(),
)



from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Literal, NamedTuple

import aiofiles
import anyio
from httpx import URL
from tqdm.asyncio import tqdm

from ..request import client as default_client
from ..typing import DownloadPostProcessMethod
from ..utils import Hash

if TYPE_CHECKING:
	from httpx import AsyncClient, Response
	from httpx._types import URLTypes


class DownloadParams(NamedTuple):
	url: 'URLTypes'
	filename: Path
	method: Literal['GET', 'POST'] = 'GET'
	md5: str | None = None


class Downloader:
	_global_client: ClassVar['AsyncClient'] = default_client

	def __init__(self, client: 'AsyncClient | None' = None, limit: int = 10):
		self._client = client or self._global_client
		self._semaphore = anyio.Semaphore(limit)

	async def _get_data(
		self,
		url: 'URLTypes',
		*,
		method: Literal['GET', 'POST'] = 'GET',
		md5: str | None = None,
	) -> bytes:
		url = URL(str(url))
		async with self._client.stream(method, url, timeout=None) as res:
			res: 'Response'
			res.raise_for_status()

			data = b''
			with tqdm(
				total=int(res.headers['content-length']),
				unit_scale=True,
				unit_divisor=1024,
				unit='B',
				desc=f'{url.path.split("/")[-1]}下载中',
				leave=False,
			) as progress_bar:
				num_bytes_downloaded = res.num_bytes_downloaded
				async for chunk in res.aiter_bytes():
					data += chunk
					progress_bar.update(res.num_bytes_downloaded - num_bytes_downloaded)
					num_bytes_downloaded = res.num_bytes_downloaded

		if md5 is not None and Hash(data).md5() != md5:
			raise

		return data

	async def download(
		self,
		url: 'URLTypes',
		filename: Path,
		*,
		method: Literal['GET', 'POST'] = 'GET',
		md5: str | None = None,
		postprocess_handler: DownloadPostProcessMethod | None = None,
		semaphore: anyio.Semaphore | None = None,
	):
		async with semaphore or self._semaphore:
			data = await self._get_data(url, method=method, md5=md5)
			if postprocess_handler is not None:
				data = postprocess_handler(data)

			filename.parent.mkdir(parents=True, exist_ok=True)
			async with aiofiles.open(filename, mode='wb') as f:
				await f.write(data)

	async def downloads(
		self,
		*params: DownloadParams,
		semaphore: anyio.Semaphore | None = None,
		progress_bar: tqdm | None = None,
		postprocess_handler: DownloadPostProcessMethod | None = None,
	) -> None:
		total = len(params)
		# 这里不能使用 or 表达式，因为 tqdm 的 total 属性还没有设置，
		# 此时调用 __bool__ 会报错
		pbar = (
			tqdm(total=total, desc='下载中', unit='file')
			if progress_bar is None
			else progress_bar
		)
		pbar.total = total
		with pbar:

			async def _handle(p: DownloadParams):
				await self.download(
					p.url,
					p.filename,
					method=p.method,
					md5=p.md5,
					postprocess_handler=postprocess_handler,
					semaphore=semaphore,
				)
				pbar.update()

			async with anyio.create_task_group() as tg:
				for param in params:
					tg.start_soon(_handle, param)



"""
字节数组读写工具类
"""

from enum import Enum
import struct


class Writer:
	"""字节写入工具类"""

	@staticmethod
	def byte(value: int) -> bytes:
		"""写入单字节"""
		return bytes([value])

	@staticmethod
	def short(value: int, little_endian: bool = True) -> bytes:
		"""写入有符号16位整数"""
		format_str = '<h' if little_endian else '>h'
		return struct.pack(format_str, value)

	@staticmethod
	def ushort(value: int, little_endian: bool = True) -> bytes:
		"""写入无符号16位整数"""
		format_str = '<H' if little_endian else '>H'
		return struct.pack(format_str, value)

	@staticmethod
	def int(value: int, little_endian: bool = True) -> bytes:
		"""写入有符号32位整数"""
		format_str = '<i' if little_endian else '>i'
		return struct.pack(format_str, value)

	@staticmethod
	def uint(value: 'int', little_endian: bool = True) -> bytes:
		"""写入无符号32位整数"""
		format_str = '<I' if little_endian else '>I'
		return struct.pack(format_str, value)

	@staticmethod
	def long(value: 'int', little_endian: bool = True) -> bytes:
		"""写入有符号64位整数"""
		format_str = '<q' if little_endian else '>q'
		return struct.pack(format_str, value)

	@staticmethod
	def ulong(value: 'int', little_endian: bool = True) -> bytes:
		"""写入无符号64位整数"""
		format_str = '<Q' if little_endian else '>Q'
		return struct.pack(format_str, value)

	@staticmethod
	def float(value: float, little_endian: bool = True) -> bytes:
		"""写入32位浮点数"""
		format_str = '<f' if little_endian else '>f'
		return struct.pack(format_str, value)

	@staticmethod
	def double(value: 'float', little_endian: bool = True) -> bytes:
		"""写入64位浮点数"""
		format_str = '<d' if little_endian else '>d'
		return struct.pack(format_str, value)

	@staticmethod
	def text(value: str) -> bytes:
		"""写入文本字符串"""
		return value.encode('utf-8')


class LengthType(Enum):
	"""长度类型枚举"""

	BYTE = 0
	UINT16 = 1
	UINT32 = 2


class GlobalLengthType:
	"""全局长度类型设置"""

	value = LengthType.BYTE


class BytesReader:
	"""字节数组读取器"""

	def __init__(
		self,
		data: bytes,
		length_type: LengthType = GlobalLengthType.value,
		little_endian: bool = True,
	) -> None:
		"""
		初始化字节读取器

		Args:
			data: 要读取的字节数据
			length_type: 字符串长度前缀类型，默认使用全局设置
			little_endian: 是否使用小端字节序，默认为True

		"""
		self.data = data
		self.offset = 0
		self.length_type = length_type
		self.little_endian = little_endian

	def seek(self, length: int, tag: str = ''):
		"""移动读取位置"""
		self.offset += length

	def set_offset(self, offset: int = 0):
		"""设置读取位置"""
		self.offset = offset

	def read(self, length: int | None = None, tag: str = '') -> bytes:
		"""读取指定长度的字节"""
		if length is not None:
			slice_data = self.data[self.offset : self.offset + length]
			self.seek(length, tag)
		else:
			# 读取到末尾
			slice_data = self.data[self.offset :]
			self.offset = len(self.data)
		return slice_data

	def text(self) -> str:
		"""读取带长度前缀的文本字符串"""
		if self.length_type == LengthType.BYTE:
			length = self.byte()
		elif self.length_type == LengthType.UINT16:
			length = self.ushort()
		elif self.length_type == LengthType.UINT32:
			length = self.uint()
		else:
			raise ValueError(f'Invalid length type: {self.length_type}')

		if length > 0:
			return self.read(length).decode('utf-8')
		else:
			return ''

	def boolean(self) -> bool:
		"""读取布尔值"""
		return bool(self.byte())

	def byte(self) -> int:
		"""读取单字节"""
		value = self.data[self.offset]
		self.offset += 1
		return value

	def short(self, little_endian: bool | None = None) -> int:
		"""读取有符号16位整数"""
		if little_endian is None:
			little_endian = self.little_endian
		data = self.read(2)
		format_str = '<h' if little_endian else '>h'
		return struct.unpack(format_str, data)[0]

	def ushort(self, little_endian: bool | None = None) -> int:
		"""读取无符号16位整数"""
		if little_endian is None:
			little_endian = self.little_endian
		data = self.read(2)
		format_str = '<H' if little_endian else '>H'
		return struct.unpack(format_str, data)[0]

	def int(self, little_endian: bool | None = None) -> int:
		"""读取有符号32位整数"""
		if little_endian is None:
			little_endian = self.little_endian
		data = self.read(4)
		format_str = '<i' if little_endian else '>i'
		return struct.unpack(format_str, data)[0]

	def uint(self, little_endian: bool | None = None) -> 'int':
		"""读取无符号32位整数"""
		if little_endian is None:
			little_endian = self.little_endian
		data = self.read(4)
		format_str = '<I' if little_endian else '>I'
		return struct.unpack(format_str, data)[0]

	def long(self, little_endian: bool | None = None) -> 'int':
		"""读取有符号64位整数"""
		if little_endian is None:
			little_endian = self.little_endian
		data = self.read(8)
		format_str = '<q' if little_endian else '>q'
		return struct.unpack(format_str, data)[0]

	def ulong(self, little_endian: bool | None = None) -> 'int':
		"""读取无符号64位整数"""
		if little_endian is None:
			little_endian = self.little_endian
		data = self.read(8)
		format_str = '<Q' if little_endian else '>Q'
		return struct.unpack(format_str, data)[0]

	def float(self, little_endian: bool | None = None) -> float:
		"""读取32位浮点数"""
		if little_endian is None:
			little_endian = self.little_endian
		data = self.read(4)
		format_str = '<f' if little_endian else '>f'
		return struct.unpack(format_str, data)[0]

	def double(self, little_endian: bool | None = None) -> 'float':
		"""读取64位浮点数"""
		if little_endian is None:
			little_endian = self.little_endian
		data = self.read(8)
		format_str = '<d' if little_endian else '>d'
		return struct.unpack(format_str, data)[0]

	def text_list(self) -> list[str]:
		"""读取带长度前缀的文本列表"""
		length = self.ushort()
		return [self.text() for _ in range(length)]

	def int_list(self) -> list['int']:
		"""读取带长度前缀的有符号32位整数列表"""
		length = self.ushort()
		return [self.int() for _ in range(length)]


# 类型定义
BytesStructSchema = list[bool | int | str | tuple | bytes | None]


class BundleOptions:
	"""打包选项"""

	def __init__(
		self,
		with_length: bool = True,
		length_type: LengthType = GlobalLengthType.value,
		little_endian: bool = True,
	):
		self.with_length = with_length
		self.length_type = length_type
		self.little_endian = little_endian


def bundle_bytes_struct(
	writer: Writer, schema: BytesStructSchema, tag: str = ''
) -> bytes:
	"""
	将结构化数据打包成字节数组

	Args:
		writer: 写入器
		schema: 数据结构模式
		tag: 标签，用于调试输出

	Returns:
		打包后的字节数组
	"""
	# 过滤掉None和undefined
	schema = [s for s in schema if s is not None]

	if tag:
		tag = f'bundle: {tag}'

	def bundle_string_with_length(value: str, length_type: LengthType) -> list[bytes]:
		"""打包带长度前缀的字符串"""
		text_bytes = writer.text(value)
		if length_type == LengthType.BYTE:
			return [writer.byte(len(text_bytes)), text_bytes]
		elif length_type == LengthType.UINT16:
			return [writer.ushort(len(text_bytes)), text_bytes]
		elif length_type == LengthType.UINT32:
			return [writer.uint(len(text_bytes)), text_bytes]

	bytes_list = []

	for v in schema:
		if isinstance(v, bool):
			bytes_list.append(writer.byte(1 if v else 0))
		elif isinstance(v, str):
			bytes_list.extend(bundle_string_with_length(v, GlobalLengthType.value))
		elif isinstance(v, int):
			bytes_list.append(writer.int(v))
		elif isinstance(v, bytes):
			bytes_list.append(v)
		elif isinstance(v, list | tuple) and len(v) >= 2 and isinstance(v[0], str):
			type_name = v[0]
			value = v[1]
			options = v[2] if len(v) > 2 else {}

			if type_name in [
				'byte',
				'short',
				'ushort',
				'int',
				'uint',
				'long',
				'ulong',
				'float',
				'double',
			]:
				little_endian = (
					options.get('littleEndian', True)
					if isinstance(options, dict)
					else True
				)
				writer_method = getattr(writer, type_name)
				if type_name == 'byte':
					bytes_list.append(writer_method(value))
				else:
					bytes_list.append(writer_method(value, little_endian))
			elif type_name == 'string':
				with_length = (
					options.get('withLength', True)
					if isinstance(options, dict)
					else True
				)
				if not with_length:
					bytes_list.append(writer.text(value))
				else:
					length_type = (
						options.get('lengthType', GlobalLengthType.value)
						if isinstance(options, dict)
						else GlobalLengthType.value
					)
					bytes_list.extend(bundle_string_with_length(value, length_type))

	# 计算总长度
	bundle_length = sum(len(b) for b in bytes_list)

	# 合并所有字节
	result = bytearray(bundle_length)
	offset = 0
	for b in bytes_list:
		result[offset : offset + len(b)] = b
		offset += len(b)

	return bytes(result)



import os

import anyio
from asyncer import syncify
import click
from tqdm.asyncio import tqdm

from albi0.update import updaters
from albi0.utils import set_directory, timer


@click.command(help='更新资源清单并下载资源文件')
@click.option(
	'-w',
	'--working-dir',
	default=None,
	type=click.Path(exists=True, file_okay=False, writable=True),
)
@click.option(
	'-n',
	'--updater-name',
	type=str,
	help='更新器名称，支持传入更新器/组名',
)
@click.option(
	'-m',
	'--manifest-path',
	type=click.Path(dir_okay=False, writable=True),
	help='本地清单文件路径，如果未指定则使用更新器默认路径',
)
@click.option(
	'-s',
	'--semaphore-limit',
	'--max-workers',
	default=10,
	type=int,
	show_default=True,
)
@click.option(
	'--version-only',
	is_flag=True,
	default=False,
	help='仅获取远程版本号，不下载资源文件',
)
@click.option(
	'--manifest-only',
	is_flag=True,
	default=False,
	help='仅更新资源清单，不下载资源文件',
)
@click.argument('patterns', nargs=-1, default=None)
@click.pass_context
@syncify
async def update(
	ctx: click.Context,
	patterns: list[str] | None,
	working_dir: str | None,
	updater_name: str,
	manifest_path: str | None,
	semaphore_limit: int,
	version_only: bool,
	manifest_only: bool,
) -> None:
	patterns = patterns or []
	os.chdir(working_dir or './')
	updater_set = updaters.get_processors(updater_name)
	if not updater_set:
		click.echo(f'找不到输入的更新器/组：{updater_name}')
		return

	_updater_string = '找到以下更新器：\n'
	_updater_string += ''.join(
		[f'    {name}: {processor.desc}\n' for name, processor in updaters.items()]
	)
	click.echo(_updater_string)

	with (
		timer('✅ 更新完成~ 总耗时: {duration:.2f}s'),
		set_directory(working_dir or './'),
	):
		for updater in updater_set:
			# 如果指定了自定义清单路径，设置到版本管理器中
			if manifest_path:
				click.echo(f'使用自定义清单文件路径: {manifest_path}')
				updater.version_manager.local_manifest_path = manifest_path

			click.echo(f'运行更新器：{updater.name}')
			if version_only:
				click.echo(
					f'远程版本：{updater.version_manager.get_remote_version()}\n'
				)
				continue

			if manifest_only:
				continue

			click.echo(
				f'本地版本：{updater.version_manager.load_local_version() or "无"}\n'
				f'远程版本：{updater.version_manager.get_remote_version()}\n'
				f'开始运行更新器...'
			)
			await updater.update(
				progress_bar=tqdm(desc='下载资源文件', unit='file'),
				patterns=patterns,
				semaphore=anyio.Semaphore(semaphore_limit),
			)
			click.echo(
				f'✅(<ゝω・)～☆更新完毕！'
				f'本地版本：{updater.version_manager.load_local_version() or "无"}\n'
			)



import click

from albi0.extract.extractor import extractors
from albi0.update.updater import updaters


@click.command(help='获取所有提取器/更新器')
@click.pass_context
def list(ctx: click.Context):
	string = '更新器组：\n'
	string += ''.join(
		[f'    {name}: {processor.desc}\n' for name, processor in updaters.items()]
	)
	string += '\n'
	string += '提取器组：\n'
	string += ''.join(
		[f'    {name}: {processor.desc}\n' for name, processor in extractors.items()]
	)
	click.echo(string)



from collections.abc import Generator, Iterable, Iterator
from contextlib import contextmanager
import fnmatch
import functools
import gzip
import hashlib
import itertools
import os
from pathlib import Path
import time
from typing import AnyStr, TypeVar

from tqdm import tqdm

from albi0.log import logger
from albi0.typing import PathTypes, T_PathLike


class Hash:
	def __init__(self, data):
		self.data = data

	@classmethod
	def from_file(cls, filename: PathTypes):
		filename = Path(filename)
		return cls(filename.read_bytes())

	def md5(self) -> str:
		# 创建一个MD5哈希对象
		return hashlib.md5(self.data).hexdigest()

	def sha256(self) -> str:
		# 创建一个SHA256哈希对象
		return hashlib.sha256(self.data).hexdigest()


class FileHash(Hash):
	def __init__(self, filename: PathTypes):
		self.filename = Path(filename)
		super().__init__(self.filename.read_bytes())


def find_files(
	filenames: Iterable[AnyStr], patterns: Iterable[AnyStr]
) -> Iterator[AnyStr]:
	if not patterns:
		for filename in filenames:
			yield filename
	else:
		filter_partial = functools.partial(fnmatch.filter, filenames)
		for filename in itertools.chain(*map(filter_partial, patterns)):
			yield filename


def decompress_file(
	filename: Path, new_filename: Path, remove_original_file: bool = False
):
	"""解压一个gzip文件"""
	with gzip.open(filename, mode='rb') as f_in:
		data = f_in.read()
	with open(new_filename, 'wb') as f_out:
		f_out.write(data)
	if remove_original_file:
		filename.unlink()


def decompress_dir(
	dirname: Path,
	*,
	pattern: str,
	new_suffix: str | None = None,
	remove_original_file: bool = False,
) -> None:
	"""解压目录下所有的gzip文件"""
	files = list(dirname.glob(pattern))
	for file in tqdm(files, desc='解压中'):
		new_filename = file
		if new_suffix is not None:
			new_filename = new_filename.with_suffix(new_suffix)
		try:
			decompress_file(file, new_filename, remove_original_file)
		except gzip.BadGzipFile:
			logger.debug(f'{file.name}不是gzip压缩文件，跳过')


def remove_all_suffixes(filename: T_PathLike) -> T_PathLike:
	"""删除路径的所有后缀"""
	root, suffix = os.path.splitext(filename)
	while suffix:
		root, suffix = os.path.splitext(root)

	return type(filename)(root)  # type: ignore


@contextmanager
def set_directory(path: PathTypes):
	"""Sets the cwd within the context

	Args:
	    path (Path): The path to the cwd

	Yields:
	    None
	"""

	origin = Path().absolute()
	try:
		os.chdir(path)
		yield
	finally:
		os.chdir(origin)


T_Path = TypeVar('T_Path', bound=PathTypes)


def join_path(path: T_Path, *args: PathTypes) -> T_Path:
	import os

	return type(path)(os.path.join(path, *args))


def join_url(base: str, *urls: str) -> str:
	from urllib.parse import urljoin

	for url in urls:
		base = urljoin(base, url)

	return base


@contextmanager
def timer(
	message_template: str = '{name} 耗时: {duration:.2f}s', **kwargs
) -> Generator[None, None, None]:
	"""一个简单的计时器上下文管理器

	Args:
		message_template: 最终展示的字符串模板.
		**kwargs: 模板中可用的额外参数.
	"""
	import click

	start_time = time.perf_counter()
	try:
		yield
	finally:
		end_time = time.perf_counter()
		duration = end_time - start_time
		click.echo(message_template.format(duration=duration, **kwargs))



from importlib.metadata import version

from asyncer import asyncify

from albi0.cli import cli

try:
	__version__ = version('albi0')
except Exception:
	__version__ = None


async def cli_main(*args, **kwargs):
	return await asyncify(cli)(*args, **kwargs)



from typing import TYPE_CHECKING

from httpx import AsyncClient
from packaging.version import Version
from UnityPy.enums.ClassIDType import ClassIDType

from albi0.bytes_reader import BytesReader
from albi0.extract.extractor import Extractor
from albi0.extract.registry import AssetPostHandlerGroup, ObjPreHandlerGroup
from albi0.typing import ObjectPath
from albi0.update import Downloader, Updater
from albi0.updaters import YooVersionManager
from albi0.updaters.yoo_version_manager import (
	PackageAssetInfo,
	PackageBundleInfo,
	YooManifestParser,
)

if TYPE_CHECKING:
	from UnityPy.classes import Texture2D


header = {
	'user-agent': r'Mozilla/5.0 (Linux; Android 6.0.1; RIDGE 4G Build/LRX22G)AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2887.55 Mobile Safari/537.36',
	'referer': r'https://newseer.61.com',
}

downloader = Downloader(AsyncClient())

obj_pre = ObjPreHandlerGroup()
asset_post = AssetPostHandlerGroup()
Extractor(
	'newseer',
	'赛尔号资源提取器',
	asset_posthandler_group=asset_post,
	obj_prehandler_group=obj_pre,
)


@obj_pre.register(ClassIDType.Sprite)
@obj_pre.register(ClassIDType.Texture2D)
def texture2d_prehandler(
	obj: 'Texture2D', obj_path: ObjectPath
) -> tuple['Texture2D', ObjectPath]:
	if obj.image.mode == 'RGBA' and obj_path.suffix != '.png':
		obj_path = obj_path.with_suffix('.png')
	return obj, obj_path


class NewseerManifestParser(YooManifestParser):
	def _parse_asset_infos(
		self,
		reader: BytesReader,
		version: str,
		count: int,
	) -> list[PackageAssetInfo]:
		"""解析资源信息列表"""
		return [
			PackageAssetInfo(
				Address='',
				AssetPath=reader.text(),
				AssetGUID=None,
				AssetTags=[],
				BundleID=reader.int(),
				DependIDs=reader.int_list(),
			)
			for _ in range(count)
		]

	def _parse_bundle_list(
		self, reader: BytesReader, version: str, count: int
	) -> list[PackageBundleInfo]:
		"""解析Bundle列表"""
		return [
			PackageBundleInfo(
				BundleName=reader.text(),
				UnityCRC=reader.uint() if Version(version) > Version('1.5.1') else None,
				FileHash=reader.text(),
				FileCRC=reader.text(),
				FileSize=reader.long(),
				IsRawFile=reader.boolean(),
				LoadMethod=reader.byte(),
				Tags=[],
				ReferenceIDs=reader.int_list(),
			)
			for _ in range(count)
		]


Updater(
	'newseer.default',
	'赛尔号AB包下载器 DefaultPackage部分',
	version_manager=YooVersionManager(
		'DefaultPackage',
		remote_path='https://newseer.61.com/Assets/StandaloneWindows64/DefaultPackage/',
		local_path='./newseer/assetbundles/DefaultPackage/',
		manifest_factory=NewseerManifestParser(),
		version_factory=int,
	),
	downloader=downloader,
)


Updater(
	'newseer.config',
	'赛尔号AB包下载器 ConfigPackage部分',
	version_manager=YooVersionManager(
		'ConfigPackage',
		remote_path='https://newseer.61.com/Assets/StandaloneWindows64/ConfigPackage/',
		local_path='./newseer/assetbundles/ConfigPackage/',
		manifest_factory=NewseerManifestParser(),
		version_factory=int,
	),
	downloader=downloader,
)


Updater(
	'newseer.pet',
	'赛尔号AB包下载器 PetAnimPackage部分',
	version_manager=YooVersionManager(
		'PetAnimPackage',
		remote_path='https://newseer.61.com/Assets/StandaloneWindows64/PetAnimPackage/',
		local_path='./newseer/assetbundles/PetAnimPackage/',
		manifest_factory=NewseerManifestParser(),
		version_factory=int,
	),
	downloader=downloader,
)


Updater(
	'newseer.startup',
	'赛尔号AB包下载器 StartupPackage部分',
	version_manager=YooVersionManager(
		'StartupPackage',
		remote_path='https://newseer.61.com/Assets/StandaloneWindows64/StartupPackage/',
		local_path='./newseer/assetbundles/StartupPackage/',
		manifest_factory=NewseerManifestParser(),
		version_factory=int,
	),
	downloader=downloader,
)



from concurrent.futures import ThreadPoolExecutor
from io import BytesIO
from pathlib import Path, PurePath
from typing import TYPE_CHECKING, Optional, TypeVar, cast

from tqdm import tqdm
from UnityPy import Environment
from UnityPy.environment import reSplit

from albi0.container import ProcessorContainer
from albi0.log import logger
from albi0.typing import (
	DecryptionMethod,
	ExportHandlerResult,
	ObjectPath,
	PathTypes,
)

from .registry import (
	AssetPostHandlerGroup,
	ExportHandlerGroup,
	ObjPreHandlerGroup,
	StopExtractThisObject,
)

if TYPE_CHECKING:
	from UnityPy.classes import PPtr
	from UnityPy.files import ObjectReader


extractors: ProcessorContainer['Extractor'] = ProcessorContainer()

_T = TypeVar('_T')


def _create_obj(obj: _T | None, type_: type[_T]) -> _T:
	return type_() if obj is None else obj


T = TypeVar('T')


def _output_as_is(x: T) -> T:
	return x


class Extractor:
	def __init__(
		self,
		name: str,
		desc: str,
		*,
		decryption_method: Optional['DecryptionMethod'] = None,
		asset_posthandler_group: Optional['AssetPostHandlerGroup'] = None,
		obj_prehandler_group: Optional['ObjPreHandlerGroup'] = None,
		export_handler_group: Optional['ExportHandlerGroup'] = None,
	) -> None:
		self.name = name
		self.desc = desc

		self.decryption_method = decryption_method or _output_as_is
		self.asset_posthandler_group = _create_obj(
			asset_posthandler_group, AssetPostHandlerGroup
		)
		self.obj_prehandler_group = _create_obj(
			obj_prehandler_group, ObjPreHandlerGroup
		)
		self.export_handler_group = _create_obj(
			export_handler_group, ExportHandlerGroup
		)

		extractors[name] = self

	def from_file_load(self, *filenames: PathTypes) -> Environment:
		env = Environment()
		for filename in tqdm(
			filenames,
			desc='加载文件到内存...',
			disable=len(filenames) == 1,
			unit='file',
		):
			if split_match := reSplit.match(str(filename)):
				name = split_match.groups()[0]
			else:
				name = filename

			with open(filename, mode='rb') as f:
				data = self.decryption_method(memoryview(f.read()))

			env.load_file(cast(BytesIO, data), name=str(name))

		return env

	def extract_asset(
		self,
		*sources: PathTypes,
		export_dir: PathTypes,
		max_workers: int = 4,
		merge_extract: bool = False,
		export_unknown_as_typetree: bool = True,
	) -> None:
		export_types_keys = list(self.export_handler_group.keys())
		export_dir = Path(export_dir)

		def defaulted_export_index(type_: 'ObjectReader'):
			try:
				return export_types_keys.index(type_.type)
			except (IndexError, ValueError):
				return 999
			except Exception as e:
				logger.opt(exception=e).error(f'{e}')
				return -1

		def handle_asset(
			environment: 'Environment',
		) -> 'list[tuple[ObjectReader, ObjectPath]]':
			_result = []
			container = sorted(
				environment.container.items(),
				key=lambda x: defaulted_export_index(x[1]),  # type: ignore
			)
			for obj_path, obj in container:
				try:
					obj = cast('PPtr', obj)
					obj = obj.deref()
					obj, _ = self.asset_posthandler_group.handle(
						obj, export_dir=export_dir
					)
					# container 中的文件名不会保留大小写，需要手动替换
					real_obj_path = PurePath(obj_path)
					if name := obj.peek_name():
						# 替换文件名但保留原有后缀
						real_obj_path = real_obj_path.with_name(
							name + real_obj_path.suffix
						)
					_result.append((obj, ObjectPath(real_obj_path)))
				except StopExtractThisObject:
					continue
				except Exception as e:
					logger.opt(exception=e).error(f'{obj_path} | {e}')
					continue

			return _result

		def export_obj(
			obj_: 'ObjectReader', obj_path_: ObjectPath
		) -> ExportHandlerResult | None:
			obj_type = obj_.type
			readed_obj_ = obj_.read()  # 返回实际对象
			if self.obj_prehandler_group:
				readed_obj_, obj_path_ = self.obj_prehandler_group.handle(
					readed_obj_, obj_path_
				)

			export_filename = Path(export_dir, obj_path_)
			export_filename.parent.mkdir(parents=True, exist_ok=True)
			# export
			return self.export_handler_group.handle(
				readed_obj_,
				obj_type,
				export_filename.with_suffix(''),
				suffix=export_filename.suffix,
				export_unknown_as_typetree=export_unknown_as_typetree,
			)

		with (
			ThreadPoolExecutor(max_workers=max_workers) as executor,
			tqdm(
				desc='提取中...',
				total=None,
				disable=not merge_extract,
				unit='object',
			) as pbar,
		):

			def export_wrap(env: Environment) -> None:
				objs = handle_asset(env)
				if not pbar.disable and pbar.total is None:
					pbar.total = len(objs)

				# 使用多进程处理每个对象的导出
				futures = [
					executor.submit(export_obj, obj, obj_path) for obj, obj_path in objs
				]
				for future in futures:
					future.add_done_callback(
						lambda _: pbar.update(1) if not pbar.disable else None
					)
					future.result()

			if merge_extract:
				env = self.from_file_load(*sources)
				export_wrap(env)
				return

			with tqdm(sources, unit='file') as not_merge_pbar:
				for source_fn in not_merge_pbar:
					not_merge_pbar.set_description(f'提取文件: {Path(source_fn).name}')
					env = self.from_file_load(source_fn)
					export_wrap(env)


Extractor('default', '默认提取器，直接提取不做任何处理')



from collections.abc import Callable
from enum import IntEnum
from pathlib import Path
from typing import Any, Protocol, TypedDict

import httpx
from packaging.version import Version

from albi0.bytes_reader import BytesReader, LengthType
from albi0.update.version import (
	AbstractVersionManager,
	LocalFileName,
	Manifest,
	ManifestItem,
)
from albi0.utils import join_path, join_url


class VersionProtocol(Protocol):
	def __init__(self, version: str):
		pass

	def __lt__(self, other: Any) -> bool:
		raise NotImplementedError

	def __le__(self, other: Any) -> bool:
		raise NotImplementedError

	def __gt__(self, other: Any) -> bool:
		raise NotImplementedError

	def __ge__(self, other: Any) -> bool:
		raise NotImplementedError

	def __eq__(self, other: object) -> bool:
		raise NotImplementedError

	def __ne__(self, other: object) -> bool:
		raise NotImplementedError


class OutputNameType(IntEnum):
	"""输出名称类型枚举"""

	HashName = 1
	BundleName_HashName = 4


class PackageAssetInfo(TypedDict):
	"""包资源信息"""

	Address: str
	AssetPath: str
	AssetGUID: str | None
	AssetTags: list[str]
	BundleID: int
	DependIDs: list[int]


class PackageBundleInfo(TypedDict):
	"""包Bundle信息"""

	BundleName: str
	UnityCRC: int | None
	FileHash: str
	FileCRC: str
	FileSize: int
	IsRawFile: bool
	LoadMethod: int
	ReferenceIDs: list[int]
	Tags: list[str]


class PackageManifest(TypedDict):
	"""包清单"""

	FileVersion: str
	EnableAddressable: bool
	LocationToLower: bool
	IncludeAssetGUID: bool
	OutputNameType: OutputNameType
	PackageName: str
	PackageVersion: str
	PackageAssetCount: int
	PackageAssetInfos: list[PackageAssetInfo]
	PackageBundleCount: int
	BundleList: list[PackageBundleInfo]


def _create_empty_manifest() -> Manifest:
	return Manifest(version='0', items={})


class YooManifestParser:
	"""清单解析器"""

	def __call__(self, data: bytes) -> PackageManifest:
		return self.parse_manifest(data)

	def parse_manifest(self, data: bytes) -> PackageManifest:
		"""解析清单数据"""
		reader = BytesReader(
			data,
			length_type=LengthType.UINT16,
			little_endian=True,
		)
		reader.uint()
		version = reader.text()
		manifest = PackageManifest(
			FileVersion=version,
			EnableAddressable=reader.boolean(),
			LocationToLower=reader.boolean()
			if Version(version) > Version('1.4.16')
			else False,
			IncludeAssetGUID=reader.boolean()
			if Version(version) > Version('1.4.16')
			else False,
			OutputNameType=OutputNameType(reader.int()),
			PackageName=reader.text(),
			PackageVersion=reader.text(),
			PackageAssetCount=0,
			PackageAssetInfos=[],
			PackageBundleCount=0,
			BundleList=[],
		)
		count = reader.int()
		manifest['PackageAssetCount'] = count
		package_asset_infos = self._parse_asset_infos(reader, version, count)
		manifest['PackageAssetInfos'] = package_asset_infos

		count = reader.int()
		manifest['PackageBundleCount'] = count
		bundle_list = self._parse_bundle_list(reader, version, count)
		manifest['BundleList'] = bundle_list
		return manifest

	def _parse_asset_infos(
		self, reader: BytesReader, version: str, count: int
	) -> list[PackageAssetInfo]:
		"""解析资源信息列表"""
		return [
			PackageAssetInfo(
				Address=reader.text(),
				AssetPath=reader.text(),
				AssetGUID=reader.text()
				if Version(version) > Version('1.4.16')
				else None,
				AssetTags=reader.text_list(),
				BundleID=reader.int(),
				DependIDs=reader.int_list(),
			)
			for _ in range(count)
		]

	def _parse_bundle_list(
		self, reader: BytesReader, version: str, count: int
	) -> list[PackageBundleInfo]:
		"""解析Bundle列表"""
		return [
			PackageBundleInfo(
				BundleName=reader.text(),
				UnityCRC=reader.uint() if Version(version) > Version('1.5.1') else None,
				FileHash=reader.text(),
				FileCRC=reader.text(),
				FileSize=reader.long(),
				IsRawFile=reader.boolean(),
				LoadMethod=reader.byte(),
				Tags=reader.text_list(),
				ReferenceIDs=reader.int_list(),
			)
			for _ in range(count)
		]


class YooVersionManager(AbstractVersionManager):
	def __init__(
		self,
		package_name: str,
		*,
		remote_path: str,
		local_path: str,
		manifest_factory: Callable[[bytes], PackageManifest] = YooManifestParser(),
		version_factory: type[VersionProtocol | float] = Version,
	) -> None:
		super().__init__()
		self.package_name = package_name
		self.remote_path = remote_path
		self.local_path = local_path
		self.manifest_factory = manifest_factory
		self.version_factory = version_factory

		self._local_manifest_path = join_path(
			self.local_path, f'PackageManifest_{self.package_name}.json'
		)
		self.version_basename = f'PackageManifest_{self.package_name}.version'

	@property
	def local_manifest_path(self) -> str:
		return self._local_manifest_path

	@local_manifest_path.setter
	def local_manifest_path(self, manifest_path: str) -> None:
		self._local_manifest_path = manifest_path

	def _simplify_manifest(self, data: PackageManifest) -> Manifest:
		"""将远程清单转换为Manifest实例。"""
		version = data['PackageVersion']
		items = {}
		for item in data['BundleList']:
			local_basename = item['BundleName']
			remote_filehash = item['FileHash']
			local_fn = LocalFileName(join_path(self.local_path, local_basename))
			items[local_fn] = ManifestItem(
				join_url(self.remote_path, remote_filehash),
				f'{local_basename}.bundle',
				remote_filehash.encode(),
			)
		return Manifest(version=version, items=items)

	def get_remote_version(self) -> str:
		"""获取远程版本号"""
		remote_version_url = join_url(self.remote_path, self.version_basename)
		result = httpx.get(remote_version_url).text
		return result

	def load_local_version(self) -> str:
		"""加载本地版本号"""
		return self.load_local_manifest().version

	def get_remote_manifest(self) -> Manifest:
		remote_manifest_url = join_url(
			self.remote_path,
			f'PackageManifest_{self.package_name}_{self.get_remote_version()}.bytes',
		)
		req = httpx.get(remote_manifest_url)
		manifest_dict = self.manifest_factory(req.content)
		return self._simplify_manifest(manifest_dict)

	def load_local_manifest(self) -> Manifest:
		if not self.is_local_version_exists:
			return _create_empty_manifest()

		return Manifest.from_json(Path(self.local_manifest_path).read_bytes())

	def save_manifest_to_local(self, manifest: Manifest) -> None:
		local_manifest_path = Path(self.local_manifest_path)
		local_manifest_path.parent.mkdir(parents=True, exist_ok=True)
		local_manifest_path.write_text(manifest.to_json())

	@property
	def is_version_outdated(self) -> bool:
		local_version = self.load_local_version() or '0'
		remote_version = self.get_remote_version() or '0'
		return not self.is_local_version_exists or (
			self.version_factory(local_version) < self.version_factory(remote_version)
		)

	@property
	def is_local_version_exists(self) -> bool:
		"""检查本地版本是否存在"""
		return Path(self.local_manifest_path).is_file()



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import NamedTuple, NewType
from typing_extensions import Self

from dataclasses_json import DataClassJsonMixin, config
import dataclasses_json.cfg

LocalFileName = NewType('LocalFileName', str)


def register_path_encoder() -> None:
	from pathlib import Path, PosixPath, WindowsPath

	for path_type in (Path, PosixPath, WindowsPath):
		dataclasses_json.cfg.global_config.encoders[path_type] = str


register_path_encoder()


class ManifestItem(NamedTuple):
	remote_filename: str
	local_basename: str
	file_hash: bytes


def encode_manifest_items(items: dict[LocalFileName, ManifestItem]) -> dict:
	return {
		str(local_fn): {
			'remote_filename': item.remote_filename,
			'local_basename': item.local_basename,
			'file_hash': item.file_hash.hex(),
		}
		for local_fn, item in items.items()
	}


def decode_manifest_items(items: dict) -> dict[LocalFileName, ManifestItem]:
	return {
		LocalFileName(local_fn): ManifestItem(
			item['remote_filename'],
			item['local_basename'],
			bytes.fromhex(item['file_hash']),
		)
		for local_fn, item in items.items()
	}


@dataclass
class Manifest(DataClassJsonMixin):
	version: str
	items: dict[LocalFileName, ManifestItem] = field(
		metadata=config(
			encoder=encode_manifest_items,
			decoder=decode_manifest_items,
		)
	)

	def filter_local_filenames_by_glob(self, *patterns: str) -> Self:
		"""使用glob模式从本地文件名中筛选

		Args:
			*patterns: 一个或多个glob模式字符串，如 "*.txt", "data/**/*.json"

		Returns:
			包含匹配文件的新Manifest实例，返回所有模式的并集，如果patterns为空，则原样返回
		"""
		from fnmatch import fnmatch

		if not patterns:
			return self

		filtered_items = {}
		for local_fn, item in self.items.items():
			for pattern in patterns:
				if fnmatch(item.local_basename, pattern):
					filtered_items[local_fn] = item
					break

		return self.__class__(version=self.version, items=filtered_items)


dataclasses_json.cfg.global_config.decoders[ManifestItem] = lambda values: ManifestItem(
	*values
)


class AbstractVersionManager(ABC):
	@property
	@abstractmethod
	def is_local_version_exists(self) -> bool:
		"""本地版本是否存在"""
		pass

	@property
	@abstractmethod
	def local_manifest_path(self) -> str:
		"""本地清单文件路径"""
		pass

	@local_manifest_path.setter
	@abstractmethod
	def local_manifest_path(self, manifest_path: str) -> None:
		"""设置自定义清单文件路径

		Args:
			manifest_path: 自定义的清单文件路径
		"""
		pass

	@abstractmethod
	def load_local_manifest(self) -> Manifest:
		pass

	@abstractmethod
	def save_manifest_to_local(self, manifest: Manifest) -> None:
		pass

	@abstractmethod
	def load_local_version(self) -> str:
		pass

	@abstractmethod
	def get_remote_manifest(self) -> Manifest:
		pass

	@abstractmethod
	def get_remote_version(self) -> str:
		pass

	@property
	@abstractmethod
	def is_version_outdated(self) -> bool:
		"""
		如果本地版本不存在或版本号小于远程版本号，返回True，反之返回False，
		注意该方法仅判断版本号，不能用于判断是否需要下载资源。
		"""

	def generate_update_manifest(self, *patterns: str) -> Manifest | None:
		"""
		比对本地与远程清单，返回需要更新的资源。

		Args:
			*patterns: 一个或多个glob模式字符串，如 "*.txt", "data/**/*.json"

		Returns:
			包含匹配资源的清单，返回所有模式的并集，patterns为空则原样返回，
			如果没有需要更新的资源，返回None。
		"""
		remote_manifest = self.get_remote_manifest()
		local_manifest = self.load_local_manifest()
		remote_version = remote_manifest.version
		remote_items = remote_manifest.filter_local_filenames_by_glob(*patterns).items
		local_items = local_manifest.filter_local_filenames_by_glob(*patterns).items

		if local_items == remote_items or not remote_items:
			return None

		def needs_update(item: tuple[LocalFileName, ManifestItem]) -> bool:
			local_fn, remote_manifest_item = item
			try:
				return local_items[local_fn].file_hash != remote_manifest_item.file_hash
			except KeyError:
				return True

		items = filter(needs_update, remote_items.items())
		if not items:
			return None

		return Manifest(version=remote_version, items=dict(items))

	def save_remote_manifest(self) -> None:
		"""保存远程资源清单到本地"""
		manifest = self.get_remote_manifest()
		self.save_manifest_to_local(manifest)



from collections.abc import Iterable
from pathlib import Path

import anyio
import click
from tqdm.asyncio import tqdm

from albi0.container import ProcessorContainer
from albi0.typing import DownloadPostProcessMethod

from .downloader import Downloader, DownloadParams
from .version import AbstractVersionManager

updaters: ProcessorContainer['Updater'] = ProcessorContainer()


class Updater:
	def __init__(
		self,
		name: str,
		desc: str,
		*,
		version_manager: AbstractVersionManager,
		downloader: Downloader,
		postprocess_handler: DownloadPostProcessMethod | None = None,
	) -> None:
		self.name = name
		self.desc = desc
		self.version_manager = version_manager
		self.downloader = downloader
		self.postprocess_handler = postprocess_handler

		updaters[self.name] = self

	def _log_message(self, message: str) -> None:
		click.echo(f'更新器|[{self.name}]: {message}')

	async def update(
		self,
		*,
		progress_bar: tqdm | None = None,
		save_manifest: bool = True,
		patterns: Iterable[str] = (),
		semaphore: anyio.Semaphore | None = None,
	) -> None:
		"""异步更新资源文件

		检查版本是否过期，如果需要更新则下载新的资源文件并保存清单，过滤后没有需要更新的文件时不会保存清单。

		Args:
			progress_bar: 进度条
			semaphore: 并发限制
			save_manifest: 是否保存清单
			patterns: glob语法的文件名过滤模式，用于过滤希望检查更新的文件，
			如果为空则检查所有文件。
		"""
		manifest = self.version_manager.generate_update_manifest(*patterns)
		if not manifest:
			self._log_message('没有需要更新的文件，运行结束')
			return

		items = manifest.items
		self._log_message(f'需要更新的文件数量: {len(items)}')
		tasks = [
			DownloadParams(url=item.remote_filename, filename=Path(local_fn))
			for local_fn, item in items.items()
		]
		if tasks:
			await self.downloader.downloads(
				*tasks,
				progress_bar=progress_bar,
				postprocess_handler=self.postprocess_handler,
				semaphore=semaphore,
			)

		if not tasks:
			self._log_message('没有需要更新的文件，不保存资源清单')
			return

		if not save_manifest:
			self._log_message('参数save_manifest为False，不保存资源清单')
			return

		self.version_manager.save_manifest_to_local(manifest)
		self._log_message('资源清单更新完成')
